// Axis-aligned bounding box class, comprised of a position P
// and extents E, where the extremes of the box are given by
// (P-E,P+E).

#ifndef __AABB_H
#define __AABB_H

#include "vec3.h"

class AABB {
	Vec3 P;
	Vec3 E;
public:
	AABB() {}
	AABB(const Vec3& p, const Vec3& e) : P(p), E(e) {}
	const Vec3& pos() const { return P; }
	const Vec3& extents() const { return E; }
	void pos(const Vec3 &v) { P = v; }
	void extents(const Vec3 &v) { E = v; }
	Vec3 min() const { return P-E; }
	Vec3 max() const { return P+E; }
};

class BSphere {
	Vec3 P;
	float r_;
public:
	BSphere() {}
	BSphere(const Vec3& p, float radius) : P(p), r_(radius) {}
	const Vec3& pos() const { return P; }
	float r() const { return r_; }
	void pos(const Vec3 &v) { P = v; }
	void r(float radius) { r_ = radius; }
};

#endif
