#include "collision.h"
#include "math.h"

// Collide an AABB and a bounding sphere, returning the resulting
// contact normal in N if a collision is detected.
bool collide(const AABB& aabb, const BSphere& s, Vec3 &N) {
	Vec3 P = s.pos() - aabb.pos();       // P: sphere center relative to box
	Vec3 Q(P);
	Vec3 E = aabb.extents();
	
	// Q: P clipped to the box
	if (P.x < -E.x) { Q.x = -E.x; }
	if (P.y < -E.y) { Q.y = -E.y; }
	if (P.z < -E.z) { Q.z = -E.z; }
	if (P.x >  E.x) { Q.x =  E.x; }
	if (P.y >  E.y) { Q.y =  E.y; }
	if (P.z >  E.z) { Q.z =  E.z; }

	// R: vector from clipped point on box to sphere center; it's
	// the distance between the points, and the contact normal.
	// If the contact point is a face, R is perpendicular (Q slides
	// along that face).  If a vertex, the angle can vary (Q is fixed).
	
	Vec3 R = P - Q;
	float length_2 = R.length_2();
	float r = s.r();
	if (length_2 > r*r) return false;         // sphere is not touching box, bail
	// float depth = r - SQRT(length_2);   // penetration depth into box along normal
	// Vec3 contact_position = aabb.pos() + Q;
	R *= RSQRT(length_2);                // normalize R
	N = R;
	return true;
}
