#ifndef __COLLISION_H
#define __COLLISION_H

#include "ball.h"
#include "brick.h"
#include "aabb.h"
#include "vec3.h"
#include "plane.h"

// This function may be worth inlining.
bool collide(const AABB& aabb, const BSphere& s, Vec3 &N);

inline bool collide(const Plane& p, const BSphere& s, Vec3 &N) {
	Vec3 n = p.N();
	float k = s.pos().dot(n);   // distance from point to hypothetical plane at origin
	float depth = s.r() - k - p.D();  // penetration depth
	if (depth < 0) return false;
	// Vec3 contact_pos = s.pos() - p.N()*s.r();
	N = n;
	return true;
}

// convenience functions
inline bool collide(const Brick& brick, const Ball& ball, Vec3 &N) {
	return collide(brick.aabb(), ball.bsphere(), N);
}

inline bool collide(const Plane& plane, const Ball& ball, Vec3 &N) {
	return collide(plane, ball.bsphere(), N);
}


#endif
