#ifndef __COLOR3_H
#define __COLOR3_H

#include <assert.h>

class Color3 {
public:
	float	r, g, b;

public:
	Color3() {}

	Color3(float ir, float ig, float ib)
		: r(ir), g(ig), b(ib) { }

	float operator[](int i) const {
		if (i == 0)
			return r;
		else if (i == 1)
			return g;
		else if (i == 2)
			return b;
		else {
			assert_msg(false, "Color3::operator[] Invalid index");
			return 0.0f;
		}
	}

	bool operator==(const Color3 &c ) const {
		return r == c.r && g == c.g && b == c.b;
	}

	bool operator!=(const Color3 &c ) const {
		return !(*this == c);
	}

	Color3 operator+(const Color3 &c) const {
		return Color3(r + c.r, g + c.g, b + c.b);
	}

	Color3 operator-(const Color3 &c) const {
		return Color3(r - c.r, g - c.g, b - c.b);
	}

	Color3 operator-() const {
		return Color3(-r, -g, -b);
	}

	Color3 operator*(float s) const {
		return Color3(r*s, g*s, b*s);
	}

	Color3 operator*(const Color3 &c) const {
		return Color3(r*c.r, g*c.g, b*c.b);
	}

	Color3& operator+=(const Color3& c) {
		r += c.r; g += c.g; b += c.b;
		return *this;
	}

	Color3 & operator-=(const Color3& c) {
		r -= c.r; g -= c.g; b -= c.b;
		return *this;
	}

	Color3 & operator*=(float s) {
		r *= s; g *= s; b *= s;
		return *this;
	}

	Color3 & operator/=(float s) {
		*this *= 1.0 / s;
		return *this;
	}

	Color3 & clamp() {
		if (r > 1.0) r = 1.0;
		if (g > 1.0) g = 1.0;
		if (b > 1.0) b = 1.0;
		return *this;
	}
};

#endif	/* __VEC4_H */

