/*
   KallistiOS ##version##
   bubbles.c

   (c)2002 Dan Potter
*/

#include <kos.h>
#include <math.h>
#include <assert.h>
#include <plx/matrix.h>
#include <plx/prim.h>
#include <vector>
#include "mat3d.h"
#include "math.h"
#include "lighting.h"
#include "color3.h"
#include "vec4.h"
#include "vec3.h"
#include "model.h"
#include "global.h"
#include "shapes.h"
#include "brick.h"
#include "ball.h"
#include "collision.h"

using std::vector;

#define prefetch(addr) asm("pref @%0" : : "r" (addr))
#define xprintf(ARGS...) if (0 && frame == 0) { printf(ARGS); }

struct Rcontext rctx;
Light *L1, *L2, *L3, *L4, *L5, *L6, *L7;

static float phase = 0.0f;
pvr_poly_cxt_t cxt;
pvr_poly_hdr_t hdr;
static int frame;
const int FPS=60;
const float one_over_FPS=1.0/FPS;

/* Mesa has four lighting paths: the first two perform full lighting using
 * every enabled light, one using separate specular -after- texturing.  The
 * other two are "fast" paths, assuming infinite (directional) lights; one
 * assumes there's only a single light.  Note that all paths always perform
 * specular calculations; even in the fast path, this will require one dot
 * product per vertex (fortunately, normalization is not required per vertex).
 * The slower paths manually check each specular light for LocalViewer mode or
 * positional status, per vertex.   */

static void draw_light_directional_only(plx_dr_state_t *dr, uint32 size, VectorF *v, 
		Vector4 *n, Vector4 *L, Color4 *diffuse) {
	float d;
	float r, g, b;
	uint32 a0;

	a0 = plx_pack_alpha(diffuse->a);
	while(size--) {
		d = dot(L, n);
		if (d > 0) {
			r = diffuse->r * d;
			g = diffuse->g * d;
			b = diffuse->b * d;
			/* Note: it's faster to do the multiplies if d==0 than to
			 * have an else {r=g=b=0} branch! */
			plx_vert_ind(dr, v->flags, v->x, v->y, v->z, a0 | plx_pack_color3(r,g,b));
		} else {
			plx_vert_ind(dr, v->flags, v->x, v->y, v->z, a0);
		}
		v++; n++;
	}
}
//#define SPHERE_CNT 23 // max 1 pos + 1 dir
#define SPHERE_CNT 1
static float r = 5;
//static Cube cube[20];
static vector<Brick> brick;
static vector<Ball> ball;
static vector<Plane> plane;

static void sphere_frame_opaque() {
	int i;
	uint64 start;
	//static sphere_t big_sphere = { 0.6f, 20, 20, NULL };
	//static sphere_t small_sphere = { 0.3f, 20, 20, NULL };
//	static sphere_t big_sphere = { 0.25f, 34, 34, NULL };
//	static sphere_t small_sphere = { 0.15f, 34, 34, NULL };
//	static sphere_t big_sphere = { 1.0f, 54, 54, NULL };
//	static sphere_t small_sphere = { 0.6f, 54, 54, NULL };
//	static sphere_t big_sphere = { 1.0f, 20, 20, NULL };
//	static sphere_t small_sphere = { 0.6f, 20, 20, NULL };
	static Sphere big_sphere(1.0, 20, 20);
	static Sphere small_sphere(0.6, 20, 20);
	static Rectangle rect(1.5,1);

	vid_border_color(255, 0, 0);
	pvr_wait_ready();
	vid_border_color(0, 255, 0);
	start = timer_ms_gettime64();

	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -25.0f);
	plx_mat3d_rotate(r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(6.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 6.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		// plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		rctx.lighting.update();
		//big_sphere.draw();
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}
	plx_mat3d_pop();

	{
		// Put attenuated Light 6 and 7 at ball 0 position.
		plx_mat3d_push();
		light_pos(*L6, Vec4(ball[0].pos(), 1));
		light_pos(*L7, Vec4(ball[1].pos(), 1));
		plx_mat3d_pop();
	}

	plx_mat3d_push();
	typedef vector<Brick>::iterator VBI;
	for (VBI it = brick.begin(); it != brick.end(); ++it) {
		Vec3 p = (*it).pos();
		plx_mat3d_translate(p.x, p.y, p.z);
		rctx.lighting.update();
		(*it).draw();
		plx_mat3d_peek();
	}
	plx_mat3d_pop();

	plx_mat3d_push();
	typedef vector<Ball>::iterator VBallI;
	for (VBallI it = ball.begin(); it != ball.end(); ++it) {
		Vec3 p = (*it).pos();
		plx_mat3d_translate(p.x, p.y, p.z);
		rctx.lighting.update();
		(*it).draw();
		plx_mat3d_peek();
	}
	plx_mat3d_pop();

	typedef vector<Plane>::iterator VPI;
	/*
	plx_mat3d_push();
	pvr_dr_init(rctx.dr);
	for (VPI it = plane.begin(); it != plane.end(); ++it) {
		Plane &plane = *it;
		Vec3 p = plane.N()*(-plane.D());
		Vec3 N = plane.N();
		plx_mat3d_translate(p.x, p.y, p.z);
		plx_mat3d_rotate(90, N.x, N.y, N.z);
		plx_vert_indm3(&rctx.dr, PVR_CMD_VERTEX, 
		plx_mat3d_peek();
	}
	plx_mat3d_pop();
	*/

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -12.0f);
	plx_mat3d_rotate(-r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(3.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 3.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		// plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		rctx.lighting.update();
		//small_sphere.draw();
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}
	plx_mat3d_pop();
	// printf("%d\n", (uint32)(timer_ms_gettime64() - start));

	vid_border_color(0, 0, 255);
	pvr_scene_finish();

	
	Vec3 N;
	for (VBallI bait = ball.begin(); bait != ball.end(); ++bait) {
		Ball& ball = *bait;  // convenient alias
		// Update ball position.
		ball.move(one_over_FPS);
		// Test collision with every brick.
		for (VBI brit = brick.begin(); brit != brick.end(); ) {
			if (collide(*brit, ball, N)) {
				// Reflect velocity according to contact normal
				   // printf("collision normal: %s\n", N.to_str().c_str());
				Vec3 V = ball.vel().reflect(N);
				ball.vel(V);
				// Remove brick
				brit = brick.erase(brit);   // ++brit
			} else {
				++brit;
			}
		}
		for (VPI plit = plane.begin(); plit != plane.end(); ++plit) {
			if (collide(*plit, ball, N)) {
				Vec3 V = ball.vel().reflect(N);
				ball.vel(V);
			}
		}
	}


	phase += 0.01;
	r = fsin(phase) * 5.0;
}

static void sphere_frame_trans() {
	int i;

	vid_border_color(255, 0, 0);
	pvr_wait_ready();
	vid_border_color(0, 255, 0);

	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_TR_POLY);

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -12.0f);
	plx_mat3d_rotate(r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(4.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 4.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		// sphere(1.0f, 20, 20);
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}

	plx_mat3d_pop();

	vid_border_color(0, 0, 255);
	pvr_scene_finish();

}

void do_sphere_test() {
	uint8		addr;
	cont_cond_t	cond;
	int		mode = 0;
	int		a_was_down = 0;

	frame = 0;

	pvr_set_bg_color(0.2f, 0.0f, 0.4f);

	pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
	cxt.gen.culling = PVR_CULLING_NONE;
	pvr_poly_compile(&hdr, &cxt);

	for (;;) {
		if (!mode) {
				sphere_frame_opaque();
		}
		else {
			sphere_frame_trans();
		}

		frame++;
		
		addr = maple_first_controller();
		if (addr == 0)
			continue;
		if (cont_get_cond(addr, &cond) < 0)
			continue;
		if (!(cond.buttons & CONT_START)) {
			return;
		}
		if (!(cond.buttons & CONT_A)) {
			if (a_was_down) continue;
			a_was_down = 1;
			mode ^= 1;
			switch(mode) {
			case 0:	/* Opaque */
				pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
				cxt.gen.culling = PVR_CULLING_NONE;
				pvr_poly_compile(&hdr, &cxt);
				break;
			case 1:	/* Translucent */
				pvr_poly_cxt_col(&cxt, PVR_LIST_TR_POLY);
				cxt.gen.culling = PVR_CULLING_NONE;
				pvr_poly_compile(&hdr, &cxt);
				break;
			}
		} else {
			a_was_down = 0;
		}
	}
}

void create_bricks() {
	// Create bricks
	const float xsep = 1.0;
	const float ysep = 1.0;
	const int XBRIX = 20;
	const int YBRIX = 10;
	const int NBRIX = XBRIX * YBRIX;
	float yo = ysep * -YBRIX/2.0;
	int n=0;

	brick.resize(NBRIX);
	for (int j=0; j < YBRIX; j++) {
		float xo = xsep * -XBRIX/2.0;
		for (int i=0; i < XBRIX; i++) {
			brick[n].init(Vec3(xo,yo,0), Vec3(0.4,0.3,0.2));
			xo += xsep;
			++n;
		}
		yo += ysep;
	}
}

void create_balls() {
	ball.reserve(2);
	// This is dangerous, because Sphere does not have a proper copy constructor.
	Ball b(Vec3(0,-7,0),0.2);
	b.vel(Vec3(2.4,4.8,0));
	ball.push_back(b);

	Ball b2(Vec3(2,-7,0),0.2);
	b2.vel(Vec3(-1.6,4.8,0));
	ball.push_back(b2);
}
	
void create_walls() {
	assert(plane.empty());
	plane.reserve(4);
	plane.push_back(Plane(Vec3(-1, 0,0), 10)); // right
	plane.push_back(Plane(Vec3( 1, 0,0), 11)); // left
	plane.push_back(Plane(Vec3( 0,-1,0), 8)); // top
	plane.push_back(Plane(Vec3( 0, 1,0), 10)); // bottom
}

pvr_init_params_t params = {
	/* Enable opaque and translucent polygons with size 16 */
	{ PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
	
	/* Vertex buffer size 512K */
	512*1024
};
	
int main(int argc, char **argv) {
	/* Init PVR API */
	if (pvr_init(&params) < 0)
		return -1;

	pvr_stats_t stats;

	Vec4 light_dir(1, 0, 0, 0);   // directional light, from the east
	Vec4 light_pos2(10, 5, 3, 1);   // positional light, in eye coordinates
	Vec4 light_pos_att(0, 0, -25, 1); // positional light, in eye coordinates
	Vec4 light_pos(0, 0, 2, 1);   // positional light, in eye coordinates
	Vec4 light_pos_4(2, 2, 2, 1);   // positional light, in eye coordinates
	Vector4 spot_direction = { 0.0f, 0.0f, -1.0f, 0.0f };   // spotlight direction, in eye coords
	Vector4 spot_direction_2 = { 0.2f, -0.1f, -1.0f, 0.0f };   // spotlight direction, in eye coords
	Vector4 spot_direction_3 = {-0.2f, +0.1f, -1.0f, 0.0f };   // spotlight direction, in eye coords
	Color3 light_color(0.0, 1.0, 0.0);
	//Color3 light_color(1.0, 1.0, 1.0);
	Color3 light_white_ambient(0.2, 0.2, 0.2);
	Color3 light_color_red(1.0, 0.2, 0.2);
	Color3 light_color_gray(0.6, 0.6, 0.6);
	Color3 light_color_blue(0.2, 0.2, 1.0);
	Color3 light_color_white(1.0, 1.0, 1.0);
	Color3 global_ambient(1, 1, 1);
	//Color3 global_ambient_null(0, 0, 0);
	Material mat;
	mat.alpha(1);
	mat.ambient(Color3(1.0, 1.0, 1.0));
	mat.diffuse(Color3(0.8, 0.2, 0.2));
	mat.specular(Color3(1, 1, 1));
	//mat.emissive(Color3(0.2, 0.2, 0.2));

	/* Init matrices */
	plx_mat3d_init();
	plx_mat3d_mode(PLX_MAT_PROJECTION);
	plx_mat3d_identity();
	plx_mat3d_perspective(45.0f, 640.0f / 480.0f, 0.1f, 100.0f);
	plx_mat3d_mode(PLX_MAT_MODELVIEW);

	init_lighting(&rctx);
	L1 = light_new(&rctx);
	validate_shine_table(&rctx, 0, 200);
	light_pos_eye(*L1, light_pos);
	//light_pos_eye(L1, &light_dir);
	light_diffuse(L1, light_color);
	light_spot_direction_eye(L1, &spot_direction);
	light_spot_exponent(L1, 100);
	light_spot_cutoff(L1, 30.0f);
	light_specular(L1, light_color);
	//light_enable_specular(L1);
	/* Combined with a white ambient material property this
	 * looks like a superbright light.  You can use specular
	 * with exponent 0 for a similar result. */
	light_ambient(L1, light_color);

	L2 = light_new(&rctx);
	light_pos_eye(*L2, light_pos2);
	//light_diffuse(L2, light_color_white);
	light_diffuse(L2, light_color_gray);
	light_specular(L2, light_color_red);
	//light_ambient(L2, light_white_ambient);
	//light_enable_specular(L2);
	
	// Directional light
	L3 = light_new(&rctx);
	light_pos_eye(*L3, light_dir);
	light_diffuse(L3, light_color_white);

	L4 = light_new(&rctx);
	light_pos_eye(*L4, light_pos);
	light_diffuse(L4, light_color_blue);
	light_ambient(L4, light_color_blue);
	light_specular(L4, light_color_blue);
	//light_enable_specular(L4);
	light_spot_direction_eye(L4, &spot_direction_2);
	light_spot_exponent(L4, 100); 
	light_spot_cutoff(L4, 20.0f);

	L5 = light_new(&rctx);
	light_pos_eye(*L5, light_pos);
	light_diffuse(L5, light_color_red);
	light_ambient(L5, light_color_red);
	light_specular(L5, light_color_red);
	//light_enable_specular(L5);
	light_spot_direction_eye(L5, &spot_direction_3);
	light_spot_exponent(L5, 100); 
	light_spot_cutoff(L5, 15.0f);

	// attenuated light
	L6 = light_new(&rctx);
	light_pos_eye(*L6, light_pos_att);
	light_ambient(L6, light_color_white);
	L6->quad_attenuation(0.6);

	// attenuated light 2
	L7 = light_new(&rctx);
	light_ambient(L7, light_color);
	L7->quad_attenuation(0.3);


	//light_enable(&rctx, L1);
	//light_enable(&rctx, L2);
	//light_enable(&rctx, L3);
	light_enable(&rctx, L4);
	//light_enable(&rctx, L5);
	light_enable(&rctx, L6);
	light_enable(&rctx, L7);
	

	rctx.lighting.set_material(mat);
	rctx.lighting.local_viewer(false);
	//rctx.lighting.global_ambient(global_ambient);
	//light_enable_global_ambient();
	
	create_bricks();
	create_balls();
	create_walls();

	/* Do the test */
	printf("Bubbles KGL sample: press START to exit, A to toggle sphere type\n");
	do_sphere_test();

	shutdown_lighting(&rctx);

    pvr_get_stats(&stats);
    dbglog(DBG_DEBUG, "3D Stats: %ld vblanks, frame rate ~%f fps, max vertex used %d bytes\n",
            stats.vbl_count, (double)stats.frame_rate, stats.vtx_buffer_used_max);


	return 0;
}

