#include <kos/vector.h>

#define MAT_FLAG_IDENTITY       0	/**< is an identity matrix flag.
					 *   (Not actually used - the identity
					 *   matrix is identified by the absense
					 /   of all other flags.) */
#define MAT_FLAG_GENERAL        0x1	/**< is a general matrix flag */
#define MAT_FLAG_ROTATION       0x2	/**< is a rotation matrix flag */
#define MAT_FLAG_TRANSLATION    0x4	/**< is a translation matrix flag */
#define MAT_FLAG_UNIFORM_SCALE  0x8	/**< is an uniform scaling matrix flag */
#define MAT_FLAG_GENERAL_SCALE  0x10	/**< is a general scaling matrix flag */
#define MAT_FLAG_GENERAL_3D     0x20	/**< general 3D matrix flag */
#define MAT_FLAG_PERSPECTIVE    0x40	/**< is a perspective projection matrix flag */
#define MAT_FLAG_SINGULAR       0x80	/**< is a singular matrix flag */
#define MAT_DIRTY_TYPE          0x100	/**< matrix type is dirty */
#define MAT_DIRTY_FLAGS         0x200	/**< matrix flags are dirty */
#define MAT_DIRTY_INVERSE       0x400	/**< matrix inverse is dirty */

/** angle preserving matrix flags mask */
#define MAT_FLAGS_ANGLE_PRESERVING (MAT_FLAG_ROTATION | \
				    MAT_FLAG_TRANSLATION | \
				    MAT_FLAG_UNIFORM_SCALE)

/** length preserving matrix flags mask */
#define MAT_FLAGS_LENGTH_PRESERVING (MAT_FLAG_ROTATION | \
				     MAT_FLAG_TRANSLATION)

/** 3D (non-perspective) matrix flags mask */
#define MAT_FLAGS_3D (MAT_FLAG_ROTATION | \
		      MAT_FLAG_TRANSLATION | \
		      MAT_FLAG_UNIFORM_SCALE | \
		      MAT_FLAG_GENERAL_SCALE | \
		      MAT_FLAG_GENERAL_3D)

/** geometry related matrix flags mask */
#define MAT_FLAGS_GEOMETRY (MAT_FLAG_GENERAL | \
			    MAT_FLAG_ROTATION | \
			    MAT_FLAG_TRANSLATION | \
			    MAT_FLAG_UNIFORM_SCALE | \
			    MAT_FLAG_GENERAL_SCALE | \
			    MAT_FLAG_GENERAL_3D | \
			    MAT_FLAG_PERSPECTIVE | \
	                    MAT_FLAG_SINGULAR)

/** dirty matrix flags mask */
#define MAT_DIRTY          (MAT_DIRTY_TYPE | \
			    MAT_DIRTY_FLAGS | \
			    MAT_DIRTY_INVERSE)

/*@}*/


/** 
 * Test geometry related matrix flags.
 * 
 * \param mat a pointer to a GLmatrix structure.
 * \param a flags mask.
 *
 * \returns non-zero if all geometry related matrix flags are contained within
 * the mask, or zero otherwise.
 */ 
#define TEST_MAT_FLAGS(flags, a)  \
    ((MAT_FLAGS_GEOMETRY & (~(a)) & (flags) ) == 0)



void transpose_4x4_matrix(matrix_t *dst, matrix_t *src);
void print_4x4_matrix(matrix_t *m, const char *msg);
void print_current_plx_matrix(const char *msg);
int invert_matrix_3d( matrix_t *out, matrix_t *in, uint32 flags );

